
/*!
 ***********************************************************************
 * \file macroblock.c
 *
 * \brief
 *     Decode a Macroblock
 *
 * \author
 *    Main contributors (see contributors.h for copyright, address and affiliation details)
 *    - Inge Lille-Langoy               <inge.lille-langoy@telenor.com>
 *    - Rickard Sjoberg                 <rickard.sjoberg@era.ericsson.se>
 *    - Jani Lainema                    <jani.lainema@nokia.com>
 *    - Sebastian Purreiter             <sebastian.purreiter@mch.siemens.de>
 *    - Thomas Wedi                     <wedi@tnt.uni-hannover.de>
 *    - Detlev Marpe                    <marpe@hhi.de>
 *    - Gabi Blaettermann               <blaetter@hhi.de>
 *    - Ye-Kui Wang                     <wyk@ieee.org>
 ***********************************************************************
*/

#include "contributors.h"

#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "global.h"
#include "mbuffer.h"
#include "elements.h"
#include "errorconcealment.h"
#include "macroblock.h"
#include "fmo.h"
#include "vlc.h"
#include "image.h"
#include "mb_access.h"

extern pic_parameter_set_rbsp_t *active_pps;
extern seq_parameter_set_rbsp_t *active_sps;
extern int  (*nal_startcode_follows) (struct img_par *img, int dummy);

static void SetMotionVectorPredictor (struct img_par  *img,
                                      int             *pmv_x,
                                      int             *pmv_y,
                                      int             ref_frame,
                                      int             list,
                                      int             ***refPic,
                                      int             ****tmp_mv,
                                      int             block_x,
                                      int             block_y,
                                      int             blockshape_x,
                                      int             blockshape_y);
/*!
 ************************************************************************
 * \brief
 *    initializes the current macroblock
 ************************************************************************
 */
void start_macroblock(struct img_par *img, int CurrentMBInScanOrder)
{
  Macroblock *currMB;   // intialization code deleted, see below, StW

  assert (img->current_mb_nr >=0 && img->current_mb_nr < img->PicSizeInMbs);

  currMB = &img->mb_data[img->current_mb_nr];

  /* Update coordinates of the current macroblock */
  img->mb_x = (img->current_mb_nr)%(img->width>>4);
  img->mb_y = (img->current_mb_nr)/(img->width>>4);
  
  /* Define vertical positions */
  img->block_y = img->mb_y <<2;      /* luma block position */
  img->pix_y   = img->mb_y <<4;   /* luma macroblock position */
  img->pix_c_y = img->mb_y <<3; /* chroma macroblock position */
  
  /* Define horizontal positions */
  img->block_x = img->mb_x <<2;      /* luma block position */
  img->pix_x   = img->mb_x <<4;   /* luma pixel position */
  img->pix_c_x = img->mb_x <<3; /* chroma pixel position */

  // Save the slice number of this macroblock. When the macroblock below
  // is coded it will use this to decide if prediction for above is possible
  currMB->slice_nr = img->current_slice_nr;
  
  CheckAvailabilityOfNeighbors();

  // Reset syntax element entries in MB struct
  currMB->qp          = img->qp ;
  currMB->mb_type     = 0;
  currMB->delta_quant = 0;
  currMB->cbp         = 0;
  currMB->cbp_blk     = 0;
  currMB->c_ipred_mode= DC_PRED_8; //GB

  memset(currMB->mvd,0,64*sizeof(int));

  // initialize img->m7 for ABT
  memset(img->m7,0,256*sizeof(int));

  // store filtering parameters for this MB 
  currMB->LFDisableIdc = img->currentSlice->LFDisableIdc;
  currMB->LFAlphaC0Offset = img->currentSlice->LFAlphaC0Offset;
  currMB->LFBetaOffset = img->currentSlice->LFBetaOffset;

}

/*!
 ************************************************************************
 * \brief
 *    set coordinates of the next macroblock
 *    check end_of_slice condition 
 ************************************************************************
 */
Boolean exit_macroblock(struct img_par *img,int eos_bit)
{
  Slice *currSlice = img->currentSlice;

  //! The if() statement below resembles the original code, which tested 
  //! img->current_mb_nr == img->PicSizeInMbs.  Both is, of course, nonsense
  //! In an error prone environment, one can only be sure to have a new
  //! picture by checking the tr of the next slice header!

  img->num_dec_mb++;
  
  if (img->num_dec_mb == img->PicSizeInMbs)
  {
    if (currSlice->next_header != EOS)
      currSlice->next_header = SOP;
    return TRUE;
  }
  // ask for last mb in the slice  UVLC
  else
  {
    img->current_mb_nr = FmoGetNextMBNr (img->current_mb_nr);

    if (img->current_mb_nr == -1)     // End of Slice group, MUST be end of slice
    {
      assert (nal_startcode_follows (img, eos_bit) == TRUE);
      return TRUE;
    }

    if(nal_startcode_follows(img, eos_bit) == FALSE) 
      return FALSE;

    if(img->cod_counter<=0)
      return TRUE;
    return FALSE;
  }
}

/*!
 ************************************************************************
 * \brief
 *    Interpret the mb mode for P-Frames
 ************************************************************************
 */
void interpret_mb_mode_P(struct img_par *img)
{
  int i;
  const int ICBPTAB[6] = {0,16,32,15,31,47};
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int         mbmode = currMB->mb_type;
  unsigned int mbmode_case;
  mbmode_case=mbmode<4 ? 1:(mbmode==4? 2:(mbmode==5 ? 3:(mbmode==6? 4:(mbmode==31 ?5:6))));
  switch(mbmode_case)
  {
  case 1:
	currMB->mb_type = mbmode;
    for (i=0;i<4;i++)	{currMB->b8mode[i] = mbmode; currMB->b8pdir[i] = 0;}
	break;
  case 2:
	currMB->mb_type = P8x8;  
    img->allrefzero = 0;
	break;
  case 3:
	currMB->mb_type = P8x8;  
    img->allrefzero = 1;
	break;
  case 4:
	currMB->mb_type = I4MB;
    for (i=0;i<4;i++)	{/*currMB->b8mode[i] = IBLOCK;*/ currMB->b8pdir[i] = -1;}
	break;
  case 5:
	currMB->mb_type=IPCM;    
    for (i=0;i<4;i++)     {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1;}
    currMB->cbp= -1;
    currMB->i16mode = 0;
	break;
  default:
	currMB->mb_type = I16MB;
    for (i=0;i<4;i++) {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; }
    currMB->cbp= ICBPTAB[(mbmode-7)>>2];
    currMB->i16mode = (mbmode-7) & 0x03;
	return;
  }
}

/*!
 ************************************************************************
 * \brief
 *    Interpret the mb mode for I-Frames
 ************************************************************************
 */
void interpret_mb_mode_I(struct img_par *img)
{
  int i;
  const int ICBPTAB[6] = {0,16,32,15,31,47};
  Macroblock *currMB   = &img->mb_data[img->current_mb_nr];
  int         mbmode   = currMB->mb_type;

  if (mbmode==0)
  {
    currMB->mb_type = I4MB;
    for (i=0;i<4;i++) {/*currMB->b8mode[i]=IBLOCK;*/ currMB->b8pdir[i]=-1; }
	return;
  }
  else if(mbmode!=25)
  {
    currMB->mb_type = I16MB;
    for (i=0;i<4;i++) {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; }

    currMB->cbp= ICBPTAB[(mbmode-1)>>2];
    currMB->i16mode = (mbmode-1) & 0x03;
	return;
  }
  else// if(mbmode==25)
  {
    currMB->mb_type=IPCM;
    for (i=0;i<4;i++) {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; }
    currMB->cbp= -1;
    currMB->i16mode = 0;
  }
}

/*!
 ************************************************************************
 * \brief
 *    init macroblock I and P frames
 ************************************************************************
 */
void init_macroblock(struct img_par *img)
{
  int i,j;

  for (i=0;i<4;i++)
  {                           // reset vectors and pred. modes
    for(j=0;j<4;j++)
    {
      dec_picture->mv[LIST_0][img->block_x+i][img->block_y+j][0]=
      dec_picture->mv[LIST_0][img->block_x+i][img->block_y+j][1]=0;

      img->ipredmode[img->block_x+i][img->block_y+j] = DC_PRED;
    }
  }

  for (j=0; j<4; j++)
    for (i=0; i<4; i++)
    {
      dec_picture->ref_idx[LIST_0][img->block_x+i][img->block_y+j] = -1;
      dec_picture->ref_pic_id[LIST_0][img->block_x+i][img->block_y+j] = INT64_MIN;
    }
}

void reset_coeffs()
{
  int i;

  // reset luma&chroma coeffs
  memset(img->cof,0,1536);//1536 = 4*6*4*4*sizeof(int)/sizeof(byte)

  int mb_nr=img->current_mb_nr;
  // CAVLC
  for (i=0; i < 4; i++)
	  memset(img->nz_coeff[mb_nr][i],0,6*sizeof(int));
}


/*!
 ************************************************************************
 * \brief
 *    Get the syntax elements from the NAL
 ************************************************************************
 */
void read_one_macroblock(struct img_par *img)
{
  int i;

  SyntaxElement currSE;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  Slice *currSlice = img->currentSlice;
  DataPartition *dP;
  int *partMap = assignSE2partition[currSlice->dp_mode];
  int  img_block_y;
//already realized in init_macroblock
//  currMB->qp = img->qp ;

  currSE.type = SE_MBTYPE;

  //  read MB mode *****************************************************************
  dP = &(currSlice->partArr[partMap[currSE.type]]);
  
  currSE.mapping = linfo_ue;

  if(img->type == I_SLICE )
  {
    //  read MB type to mb_type
    dP->readSyntaxElement(&currSE,dP);

    currMB->mb_type = currSE.value1;
    if(!dP->bitstream->ei_flag)
      currMB->ei_flag = 0;

	interpret_mb_mode_I(img);

	read_ipred_modes(img);
	readCBPandCoeffsFromNAL (img);
	return;
  }

  // VLC Non-Intra
  else
  {
    if(img->cod_counter == -1)
    {
      //mb_skip_run
      dP->readSyntaxElement(&currSE,dP);
      img->cod_counter = currSE.value1;
    }
    if (img->cod_counter==0)
    {
      // read MB type to mb_type
      dP->readSyntaxElement(&currSE,dP);
      currSE.value1++; //P-Frame
      currMB->mb_type = currSE.value1;
      if(!dP->bitstream->ei_flag)
        currMB->ei_flag = 0;
      img->cod_counter--;
    } 
    else
    {
      img->cod_counter--;
      currMB->mb_type = 0;
      currMB->ei_flag = 0;
    }

	interpret_mb_mode_P(img);

	//====== READ 8x8 SUB-PARTITION MODES (modes of 8x8 blocks) and Intra VBST block modes ======
	if (IS_P8x8 (currMB))
	{
	 currSE.type    = SE_MBTYPE;
	 dP = &(currSlice->partArr[partMap[SE_MBTYPE]]);
	 static const int p_v2b8 [ 5] = {4, 5, 6, 7, IBLOCK};
	 static const int p_v2pd [ 5] = {0, 0, 0, 0, -1};
	 for (i=0; i<4; i++)
	 {
		 currSE.mapping = linfo_ue;
	     //sub_mb_type
	     dP->readSyntaxElement (&currSE, dP);
		 //Sets mode for 8x8 block
		 currMB->b8mode[i]   = p_v2b8[currSE.value1];
		 currMB->b8pdir[i]   = p_v2pd[currSE.value1]; 
	 }
	}

	if(img->constrained_intra_pred_flag && (img->type==P_SLICE))        // inter frame
	{
		if( !IS_INTRA(currMB) )
			img->intra_block[img->current_mb_nr] = 0;
	}
  }

  //! TO for Error Concelament
  //! If we have an INTRA Macroblock and we lost the partition
  //! which contains the intra coefficients Copy MB would be better 
  //! than just a grey block.
  //! Seems to be a bit at the wrong place to do this right here, but for this case 
  //! up to now there is no other way.
  dP = &(currSlice->partArr[partMap[SE_CBP_INTRA]]);
  if(IS_INTRA (currMB) && dP->bitstream->ei_flag && img->number)
  {
    currMB->mb_type = 0;
    currMB->ei_flag = 1;
    for (i=0;i<4;i++) {currMB->b8mode[i]=currMB->b8pdir[i]=0; }
  }
  dP = &(currSlice->partArr[partMap[currSE.type]]);
  //! End TO


  //--- init macroblock data ---
  init_macroblock       (img);

  if (IS_COPY (currMB)) //keep last macroblock
  {
    int i, j, pmv[2];
    int zeroMotionAbove;
    int zeroMotionLeft;
    PixelPos mb_a, mb_b;
    int      a_mv_y = 0;
    int      a_ref_idx = 0;
    int      b_mv_y = 0;
    int      b_ref_idx = 0;

    getLuma4x4Neighbour(img->current_mb_nr,0,0,-1, 0,&mb_a);
    getLuma4x4Neighbour(img->current_mb_nr,0,0, 0,-1,&mb_b);

    if (mb_a.available)
    {
      a_mv_y    = dec_picture->mv[LIST_0][mb_a.pos_x][mb_a.pos_y][1];
      a_ref_idx = dec_picture->ref_idx[LIST_0][mb_a.pos_x][mb_a.pos_y];
    }

    if (mb_b.available)
    {
      b_mv_y    = dec_picture->mv[LIST_0][mb_b.pos_x][mb_b.pos_y][1];
      b_ref_idx = dec_picture->ref_idx[LIST_0][mb_b.pos_x][mb_b.pos_y];
    }
    
    zeroMotionLeft  = !mb_a.available ? 1 : a_ref_idx==0 && dec_picture->mv[LIST_0][mb_a.pos_x][mb_a.pos_y][0]==0 && a_mv_y==0 ? 1 : 0;
    zeroMotionAbove = !mb_b.available ? 1 : b_ref_idx==0 && dec_picture->mv[LIST_0][mb_b.pos_x][mb_b.pos_y][0]==0 && b_mv_y==0 ? 1 : 0;

    currMB->cbp = 0;
    reset_coeffs();

    img_block_y   = img->block_y;
  
    if (zeroMotionAbove || zeroMotionLeft)
    {
      for(i=0;i<4;i++)
		  for(j=0;j<4;j++){
            dec_picture->mv[LIST_0][img->block_x+i][img_block_y+j][0] = 0;
            dec_picture->mv[LIST_0][img->block_x+i][img_block_y+j][1] = 0;
			dec_picture->ref_idx[LIST_0][img->block_x+i][img_block_y+j] = 0;
			dec_picture->ref_pic_id[LIST_0][img->block_x+i][img_block_y+j] 
				= dec_picture->ref_pic_num[LIST_0][dec_picture->ref_idx[LIST_0][img->block_x+i][img_block_y+j]];
		  }
    }
    else
    {
      SetMotionVectorPredictor (img, pmv, pmv+1, 0, LIST_0, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16, 16);
      
      for(i=0;i<4;i++)
        for(j=0;j<4;j++)
          {
            dec_picture->mv[LIST_0][img->block_x+i][img_block_y+j][0] = pmv[0];
			dec_picture->mv[LIST_0][img->block_x+i][img_block_y+j][1] = pmv[1];
			dec_picture->ref_idx[LIST_0][img->block_x+i][img_block_y+j] = 0;
			dec_picture->ref_pic_id[LIST_0][img->block_x+i][img_block_y+j] 
				= dec_picture->ref_pic_num[LIST_0][dec_picture->ref_idx[LIST_0][img->block_x+i][img_block_y+j]];
          }
    }
    return;
  }
  if(currMB->mb_type!=IPCM)
  {    
    // intra prediction modes for a macroblock 4x4 **********************************************
    read_ipred_modes(img);    
    // read inter frame vector data *********************************************************
    readMotionInfoFromNAL (img);
    // read CBP and Coeffs  ***************************************************************
    readCBPandCoeffsFromNAL (img);
	return;
  }
  else
  {
    //read pcm_alignment_zero_bit and pcm_byte[i] 
    
    // here dP is assigned with the same dP as SE_MBTYPE, because IPCM syntax is in the 
    // same category as MBTYPE
    dP = &(currSlice->partArr[partMap[SE_MBTYPE]]);
    readIPCMcoeffsFromNAL(img,dP);
  }
}

/*!
 ************************************************************************
 * \brief
 *    Read IPCM pcm_alignment_zero_bit and pcm_byte[i] from stream to img->cof
 *    (for IPCM CABAC and IPCM CAVLC  28/11/2003)
 *
 * \author
 *    Dong Wang <Dong.Wang@bristol.ac.uk>
 ************************************************************************
 */

void readIPCMcoeffsFromNAL(struct img_par *img, struct datapartition *dP)
{
  SyntaxElement currSE;
  unsigned int i,j;
  
  //read bits to let stream byte aligned
  i=(dP->bitstream->frame_bitoffset)&0x0111;
  if(i!=0)
  {
    currSE.len=8-i;
    readSyntaxElement_FLC(&currSE, dP->bitstream);
  }

  //read luma and chroma IPCM coefficients
  currSE.len=8;

  for(i=0;i<16;i++)
    for(j=0;j<16;j++)
    {
      readSyntaxElement_FLC(&currSE, dP->bitstream);
      img->cof[i>>2][j>>2][i&0x0003][j&0x0003]=currSE.value1;
    }

  for(i=0;i<8;i++)
    for(j=0;j<8;j++)
    {
      readSyntaxElement_FLC(&currSE, dP->bitstream);
      img->cof[i>>2][j>>2+4][i&0x0003][j&0x0003]=currSE.value1;
    }

  for(i=0;i<8;i++)
    for(j=0;j<8;j++)
    {
      readSyntaxElement_FLC(&currSE, dP->bitstream);
      img->cof[i>>2+2][j>>2+4][i&0x0003][j&0x0003]=currSE.value1;
    }
}



void read_ipred_modes(struct img_par *img)
{
  int b8,i,j,bi,bj,bx,by,dec;
  SyntaxElement currSE;
  Slice *currSlice;
  DataPartition *dP;
  int *partMap;
  Macroblock *currMB;
  int mostProbableIntraPredMode;
  int upIntraPredMode;
  int leftIntraPredMode;
  int IntraChromaPredModeFlag;

  PixelPos left_block;
  PixelPos top_block;
  
  currMB = &img->mb_data[img->current_mb_nr];

  IntraChromaPredModeFlag = IS_INTRA(currMB);

  currSlice = img->currentSlice;
  partMap = assignSE2partition[currSlice->dp_mode];

  currSE.type = SE_INTRAPREDMODE;

  //intra4x4_pred_mode
  dP = &(currSlice->partArr[partMap[currSE.type]]);

  if(currMB->mb_type == I4MB)
  {
    for(b8=0;b8<4;b8++)  //loop 8x8 blocks
	{
//   IntraChromaPredModeFlag = 1;

      for(j=0;j<2;j++)  //loop subblocks
        for(i=0;i<2;i++)
        {
          //get from stream
          readSyntaxElement_Intra4x4PredictionMode(&currSE,dP);

          bx = ((b8&1)<<1) + i;
          by = (b8&2)      + j;

          getLuma4x4Neighbour(img->current_mb_nr, bx, by, -1,  0, &left_block);
          getLuma4x4Neighbour(img->current_mb_nr, bx, by,  0, -1, &top_block);
          
          //get from array and decode
          bi = img->block_x + bx;
          bj = img->block_y + by;

          if (img->constrained_intra_pred_flag)
          {
            left_block.available = left_block.available ? img->intra_block[left_block.mb_addr] : 0;
            top_block.available  = top_block.available  ? img->intra_block[top_block.mb_addr]  : 0;
          }

          upIntraPredMode            = (top_block.available) ? img->ipredmode[top_block.pos_x ][top_block.pos_y ] : -1;
          leftIntraPredMode          = (left_block.available) ? img->ipredmode[left_block.pos_x][left_block.pos_y] : -1;

          mostProbableIntraPredMode  = (upIntraPredMode < 0 || leftIntraPredMode < 0) ? DC_PRED : upIntraPredMode < leftIntraPredMode ? upIntraPredMode : leftIntraPredMode;

          dec = (currSE.value1 == -1) ? mostProbableIntraPredMode : currSE.value1 + (currSE.value1 >= mostProbableIntraPredMode);

          //set
          img->ipredmode[bi][bj]=dec;
        }
    }
  }

  if (IntraChromaPredModeFlag)
  {
    currSE.type = SE_INTRAPREDMODE;
    //intra_chroma_pred_mode
    dP = &(currSlice->partArr[partMap[currSE.type]]);

    currSE.mapping = linfo_ue;

    dP->readSyntaxElement(&currSE,dP);
    currMB->c_ipred_mode = currSE.value1;

    if (currMB->c_ipred_mode < DC_PRED_8 || currMB->c_ipred_mode > PLANE_8)
    {
      error("illegal chroma intra pred mode!\n", 600);
    }
  }
}



/*!
 ************************************************************************
 * \brief
 *    Set motion vector predictor
 ************************************************************************
 */
static void SetMotionVectorPredictor (struct img_par  *img,
                                      int             *pmv_x,
                                      int             *pmv_y,
                                      int             ref_frame,
                                      int             list,
                                      int             ***refPic,
                                      int             ****tmp_mv,
                                      int             block_x,
                                      int             block_y,
                                      int             blockshape_x,
                                      int             blockshape_y)
{
  int mb_x                 = 4*block_x;
  int mb_y                 = 4*block_y;
  int mb_nr                = img->current_mb_nr;

  int mv_a, mv_b, mv_c, pred_vec=0;
  int mvPredType, rFrameL, rFrameU, rFrameUR;
  int hv;


  PixelPos block_a, block_b, block_c, block_d;

  getLuma4x4Neighbour(mb_nr, block_x, block_y,           -1,  0, &block_a);
  getLuma4x4Neighbour(mb_nr, block_x, block_y,            0, -1, &block_b);
  getLuma4x4Neighbour(mb_nr, block_x, block_y, blockshape_x, -1, &block_c);
  getLuma4x4Neighbour(mb_nr, block_x, block_y,           -1, -1, &block_d);

  if (mb_y > 0)
  {
    if (mb_x < 8)  // first column of 8x8 blocks
    {
      if (mb_y==8)
      {
        if (blockshape_x == 16)      block_c.available  = 0;
        else                         block_c.available &= 1;
      }
      else
      {
        if (mb_x+blockshape_x != 8)  block_c.available &= 1;
        else                         block_c.available  = 0;
      }
    }
    else
    {
      if (mb_x+blockshape_x != 16)   block_c.available &= 1;
      else                           block_c.available  = 0;
    }
  }

  if (!block_c.available)
  {
    block_c=block_d;
  }

  mvPredType = MVPRED_MEDIAN;

  rFrameL    = block_a.available    ? refPic[list][block_a.pos_x][block_a.pos_y] : -1;
  rFrameU    = block_b.available    ? refPic[list][block_b.pos_x][block_b.pos_y] : -1;
  rFrameUR   = block_c.available    ? refPic[list][block_c.pos_x][block_c.pos_y] : -1;

  /* Prediction if only one of the neighbors uses the reference frame
   * we are checking
   */
  if(rFrameL == ref_frame && rFrameU != ref_frame && rFrameUR != ref_frame)       mvPredType = MVPRED_L;
  else if(rFrameL != ref_frame && rFrameU == ref_frame && rFrameUR != ref_frame)  mvPredType = MVPRED_U;
  else if(rFrameL != ref_frame && rFrameU != ref_frame && rFrameUR == ref_frame)  mvPredType = MVPRED_UR;
  // Directional predictions 
  if(blockshape_x == 8 && blockshape_y == 16)
  {
	  if((mb_x == 0) && (rFrameL == ref_frame))
		  mvPredType = MVPRED_L;
	  else if((mb_x != 0) && (rFrameUR == ref_frame))
		  mvPredType = MVPRED_UR;
  }
  else if(blockshape_x == 16 && blockshape_y == 8)
  {
    if((mb_y == 0)&& (rFrameU == ref_frame))
        mvPredType = MVPRED_U;
    else if((mb_y != 0)&& (rFrameL == ref_frame))
        mvPredType = MVPRED_L;
  }

  for (hv=0; hv < 2; hv++)
  {
    mv_a = block_a.available  ? tmp_mv[list][block_a.pos_x][block_a.pos_y][hv] : 0;
    mv_b = block_b.available  ? tmp_mv[list][block_b.pos_x][block_b.pos_y][hv] : 0;
    mv_c = block_c.available  ? tmp_mv[list][block_c.pos_x][block_c.pos_y][hv] : 0;
    
    switch (mvPredType)
    {
    case MVPRED_MEDIAN:
      if(!(block_b.available || block_c.available))
        pred_vec = mv_a;
      else
        pred_vec = mv_a+mv_b+mv_c-min(mv_a,min(mv_b,mv_c))-max(mv_a,max(mv_b,mv_c));
      break;
    case MVPRED_L:
      pred_vec = mv_a;
      break;
    case MVPRED_U:
      pred_vec = mv_b;
      break;
    case MVPRED_UR:
      pred_vec = mv_c;
      break;
    default:
      break;
    }

    if (hv==0)  *pmv_x = pred_vec;
    else        *pmv_y = pred_vec;

  }
}

/*!
 ************************************************************************
 * \brief
 *    Read motion info
 ************************************************************************
 */
void readMotionInfoFromNAL (struct img_par *img)
{
  int i,j,k;
  int step_h,step_v;
  int curr_mvd;
  Macroblock *currMB  = &img->mb_data[img->current_mb_nr];
  SyntaxElement currSE;
  Slice *currSlice    = img->currentSlice;
  DataPartition *dP;
  int *partMap        = assignSE2partition[currSlice->dp_mode];
  int partmode        = (IS_P8x8(currMB)?4:currMB->mb_type);
  int step_h0         = BLOCK_STEP [partmode][0];
  int step_v0         = BLOCK_STEP [partmode][1];

  int mv_mode, i0, j0, refframe;
  int pmv[2];
  int j4, i4, ii,jj;
  int vec;

  int flag_mode;

  //  If multiple ref. frames, read reference frame for the MB *********************************
  if(img->num_ref_idx_l0_active>1) 
  {
    flag_mode = ( img->num_ref_idx_l0_active == 2 ? 1 : 0);

    currSE.type = SE_REFFRAME;
    dP = &(currSlice->partArr[partMap[SE_REFFRAME]]);

    currSE.mapping = linfo_ue;
    
    for (j0=0; j0<4; j0+=step_v0)
    {
      for (i0=0; i0<4; i0+=step_h0)
      {
        k=2*(j0/2)+(i0/2);
        if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
        {
          //ref_idx_l0

          if (!IS_P8x8 (currMB) || (!img->allrefzero))
          {
            if( (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag) && flag_mode )
            {
              currSE.len = 1;
              readSyntaxElement_FLC(&currSE, dP->bitstream);
              currSE.value1 = 1 - currSE.value1;
            }
            else
            {
              currSE.value2 = LIST_0;
              dP->readSyntaxElement (&currSE,dP);
            }
            refframe = currSE.value1;
            
          }
          else
          {
            refframe = 0;
          }
          
          for (j=j0; j<j0+step_v0;j++)
            for (i=i0; i<i0+step_h0;i++)
            {
              dec_picture->ref_idx[LIST_0][img->block_x + i][img->block_y + j] = refframe;
            }
          
        }
      }
    }
  }
  else
  {
    for (j0=0; j0<4; j0+=step_v0)
    {
      for (i0=0; i0<4; i0+=step_h0)
      {
        k=2*(j0/2)+(i0/2);
        if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
        {
          for (i=i0; i<i0+step_h0;i++)
            for (j=j0; j<j0+step_v0;j++)
              dec_picture->ref_idx[LIST_0][img->block_x + i][img->block_y + j] = 0;
        }
      }
    }
  }

  //=====  READ FORWARD MOTION VECTORS =====
  currSE.type = SE_MVD;
  dP = &(currSlice->partArr[partMap[SE_MVD]]);

  currSE.mapping = linfo_se;

  for (j0=0; j0<4; j0+=step_v0)
    for (i0=0; i0<4; i0+=step_h0)
    {
      k=2*(j0/2)+(i0/2);

      if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && (currMB->b8mode[k] !=0))//has forward vector
      {
        mv_mode  = currMB->b8mode[k];
        step_h   = BLOCK_STEP [mv_mode][0];
        step_v   = BLOCK_STEP [mv_mode][1];
        
        refframe = dec_picture->ref_idx[LIST_0][img->block_x+i0][img->block_y+j0];
        
        for (j=j0; j<j0+step_v0; j+=step_v)
        {
          for (i=i0; i<i0+step_h0; i+=step_h)
          {
            j4 = img->block_y+j;
            i4 = img->block_x+i;
            
            // first make mv-prediction
            SetMotionVectorPredictor (img, pmv, pmv+1, refframe, LIST_0, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v);

            for (k=0; k < 2; k++) 
            {
              //mvd_l0
              currSE.value2 = k<<1; // identifies the component; only used for context determination
              dP->readSyntaxElement(&currSE,dP);
              curr_mvd = currSE.value1; 
              
              vec=curr_mvd+pmv[k];           /* find motion vector */
              
              for(ii=0;ii<step_h;ii++)
              {
                for(jj=0;jj<step_v;jj++)
                {
                  dec_picture->mv  [LIST_0][i4+ii][j4+jj][k] = vec;
                  currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd;
                }
              }
            }
          }
        }
      }
  }

  // record reference picture Ids for deblocking decisions
 
  for(i4=img->block_x;i4<(img->block_x+4);i4++)
	 for(j4=img->block_y;j4<(img->block_y+4);j4++)
	 {
	   if(dec_picture->ref_idx[LIST_0][i4][j4]>=0)
	      dec_picture->ref_pic_id[LIST_0][i4][j4] = dec_picture->ref_pic_num[LIST_0][dec_picture->ref_idx[LIST_0][i4][j4]];
	  else
	     dec_picture->ref_pic_id[LIST_0][i4][j4] = INT64_MIN;
  }
}



/*!
 ************************************************************************
 * \brief
 *    Get the Prediction from the Neighboring Blocks for Number of Nonzero Coefficients 
 *    
 *    Luma Blocks
 ************************************************************************
 */
int predict_nnz(struct img_par *img, int i,int j)
{
  PixelPos pix;

  int pred_nnz = 0;
  int cnt      = 0;
  int mb_nr    = img->current_mb_nr;

  // left block
  getLuma4x4Neighbour(mb_nr, i, j, -1, 0, &pix);

  if (pix.available)
  {
    pred_nnz = img->nz_coeff [pix.mb_addr ][pix.x][pix.y];
    cnt++;
  }

  // top block
  getLuma4x4Neighbour(mb_nr, i, j, 0, -1, &pix);

  if (pix.available)
  {
    pred_nnz += img->nz_coeff [pix.mb_addr ][pix.x][pix.y];
    cnt++;
  }

  if (cnt==2)
  {
    pred_nnz++;
    pred_nnz/=cnt; 
  }

  return pred_nnz;
}


/*!
 ************************************************************************
 * \brief
 *    Get the Prediction from the Neighboring Blocks for Number of Nonzero Coefficients 
 *    
 *    Chroma Blocks   
 ************************************************************************
 */
int predict_nnz_chroma(struct img_par *img, int i,int j)
{
  PixelPos pix;

  int pred_nnz = 0;
  int cnt      =0;
  int mb_nr    = img->current_mb_nr;

  // left block
  getChroma4x4Neighbour(mb_nr, i&0x0001, j-4, -1, 0, &pix);

  if (pix.available)
  {
    pred_nnz = img->nz_coeff [pix.mb_addr ][2 * (i/2) + pix.x][4 + pix.y];
    cnt++;
  }
  
  // top block
  getChroma4x4Neighbour(mb_nr, i&0x0001, j-4, 0, -1, &pix);

  if (pix.available)
  {
    pred_nnz += img->nz_coeff [pix.mb_addr ][2 * (i/2) + pix.x][4 + pix.y];
    cnt++;
  }

  if (cnt==2)
  {
    pred_nnz++;
    //pred_nnz/=cnt; 
	pred_nnz>>=1;
  }

  return pred_nnz;
}


/*!
 ************************************************************************
 * \brief
 *    Reads coeff of an 4x4 block (CAVLC)
 *
 * \author
 *    Karl Lillevold <karll@real.com>
 *    contributions by James Au <james@ubvideo.com>
 ************************************************************************
 */


void readCoeff4x4_CAVLC (struct img_par *img,
                        int block_type, 
                        int i, int j, int levarr[16], int runarr[16],
                        int *number_coefficients)
{
  int mb_nr = img->current_mb_nr;
  Macroblock *currMB = &img->mb_data[mb_nr];
  SyntaxElement currSE;
  Slice *currSlice = img->currentSlice;
  DataPartition *dP;
  int *partMap = assignSE2partition[currSlice->dp_mode];


  int k, code, vlcnum;
  int numcoeff, numtrailingones, numcoeff_vlc;
  int level_two_or_higher;
  int numones, totzeros, level, cdc=0, cac=0;
  int zerosleft, ntr, dptype = 0;
  int max_coeff_num = 0, nnz;
  char type[15];
  int incVlc[] = {0,3,6,12,24,48,32768};    // maximum vlc = 6

  numcoeff = 0;

  switch (block_type)
  {
  case LUMA:
    max_coeff_num = 16;
    sprintf(type, "%s", "Luma");
    if (IS_INTRA (currMB))
    {
      dptype = SE_LUM_AC_INTRA;
    }
    else
    {
      dptype = SE_LUM_AC_INTER;
    }
    break;
  case LUMA_INTRA16x16DC:
    max_coeff_num = 16;
    sprintf(type, "%s", "Lum16DC");
    dptype = SE_LUM_DC_INTRA;
    break;
  case LUMA_INTRA16x16AC:
    max_coeff_num = 15;
    sprintf(type, "%s", "Lum16AC");
    dptype = SE_LUM_AC_INTRA;
    break;

  case CHROMA_DC:
    max_coeff_num = 4;
    cdc = 1;

    sprintf(type, "%s", "ChrDC");
    if (IS_INTRA (currMB))
    {
      dptype = SE_CHR_DC_INTRA;
    }
    else
    {
      dptype = SE_CHR_DC_INTER;
    }
    break;
  case CHROMA_AC:
    max_coeff_num = 15;
    cac = 1;
    sprintf(type, "%s", "ChrAC");
    if (IS_INTRA (currMB))
    {
      dptype = SE_CHR_AC_INTRA;
    }
    else
    {
      dptype = SE_CHR_AC_INTER;
    }
    break;
  default:
    error ("readCoeff4x4_CAVLC: invalid block type", 600);
    break;
  }

  currSE.type = dptype;
  dP = &(currSlice->partArr[partMap[dptype]]);

  img->nz_coeff[img->current_mb_nr][i][j] = 0;


  if (!cdc)
  {
    // luma or chroma AC
    if (!cac)
    {
      nnz = predict_nnz(img, i, j);
    }
    else
    {
      nnz = predict_nnz_chroma(img, i, j);
    }

	numcoeff_vlc = (nnz < 2)? 0:((nnz < 4)? 1:((nnz < 8)? 2:3));

    currSE.value1 = numcoeff_vlc;

    readSyntaxElement_NumCoeffTrailingOnes(&currSE, dP, type);

    numcoeff =  currSE.value1;
    numtrailingones =  currSE.value2;

    img->nz_coeff[img->current_mb_nr][i][j] = numcoeff;
  }
  else
  {
    // chroma DC
    readSyntaxElement_NumCoeffTrailingOnesChromaDC(&currSE, dP);

    numcoeff =  currSE.value1;
    numtrailingones =  currSE.value2;
  }

  memset(levarr,0,max_coeff_num*4);
  memset(runarr,0,max_coeff_num*4);

  numones = numtrailingones;
  *number_coefficients = numcoeff;

  if (numcoeff)
  {
    if (numtrailingones)
    {

      currSE.len = numtrailingones;

      readSyntaxElement_FLC (&currSE, dP->bitstream);

      code = currSE.inf;
      ntr = numtrailingones;
      for (k = numcoeff-1; k > numcoeff-1-numtrailingones; k--)
      {
        ntr --;
        if ((code>>ntr)&1)
          levarr[k] = -1;
        else
          levarr[k] = 1;
      }
    }

    // decode levels
    level_two_or_higher = 1;
    if (numcoeff > 3 && numtrailingones == 3)
      level_two_or_higher = 0;

      if (numcoeff > 10 && numtrailingones < 3)
          vlcnum = 1;
      else
          vlcnum = 0;

    for (k = numcoeff - 1 - numtrailingones; k >= 0; k--)
    {
      if (vlcnum == 0)
          readSyntaxElement_Level_VLC0(&currSE, dP);
      else
          readSyntaxElement_Level_VLCN(&currSE, vlcnum, dP);

      if (level_two_or_higher)
      {
          if (currSE.inf > 0)
          currSE.inf ++;
          else
          currSE.inf --;
          level_two_or_higher = 0;
      }

      level = levarr[k] = currSE.inf;
      if (abs(level) == 1)
        numones ++;

      // update VLC table
      if (abs(level)>incVlc[vlcnum])
        vlcnum++;

      if (k == numcoeff - 1 - numtrailingones && abs(level)>3)
        vlcnum = 2;

    }
    
    if (numcoeff < max_coeff_num)
    {
      // decode total run
      vlcnum = numcoeff-1;
      currSE.value1 = vlcnum;

      if (cdc)
        readSyntaxElement_TotalZerosChromaDC(&currSE, dP);
      else
        readSyntaxElement_TotalZeros(&currSE, dP);

      totzeros = currSE.value1;
    }
    else
    {
      totzeros = 0;
    }

    // decode run before each coefficient
    zerosleft = totzeros;
    i = numcoeff-1;
    if (zerosleft > 0 && i > 0)
    {
      do 
      {
        // select VLC for runbefore
        vlcnum = zerosleft - 1;
        if (vlcnum > RUNBEFORE_NUM-1)
          vlcnum = RUNBEFORE_NUM-1;

        currSE.value1 = vlcnum;

        readSyntaxElement_Run(&currSE, dP);
        runarr[i] = currSE.value1;

        zerosleft -= runarr[i];
        i --;
      } while (zerosleft != 0 && i != 0);
    }
    runarr[i] = zerosleft;

  } // if numcoeff
}



/*!
 ************************************************************************
 * \brief
 *    Get coded block pattern and coefficients (run/level)
 *    from the NAL
 ************************************************************************
 */
void readCBPandCoeffsFromNAL(struct img_par *img)
{
  int i,j,k;
  int level;
  int mb_nr = img->current_mb_nr;
  int ii,jj;
  int i1,j1, m2,jg2;
  Macroblock *currMB = &img->mb_data[mb_nr];
  int cbp;
  SyntaxElement currSE;
  Slice *currSlice = img->currentSlice;
  DataPartition *dP;
  int *partMap = assignSE2partition[currSlice->dp_mode];
  int coef_ctr, i0, j0, b8;
  int ll;
  int block_x,block_y;
  int start_scan;
  int uv;
  int qp_uv;
//  int run, len;
  int levarr[16], runarr[16], numcoeff;

  int qp_per    = (img->qp-MIN_QP)/6;
  int qp_rem    = (img->qp-MIN_QP)%6;
  int qp_per_uv = QP_SCALE_CR[img->qp-MIN_QP]/6;
  int qp_rem_uv = QP_SCALE_CR[img->qp-MIN_QP]%6;

  // QPI
  qp_uv = img->qp + active_pps->chroma_qp_index_offset;
  qp_uv = Clip3(0, 51, qp_uv);
  qp_per_uv = QP_SCALE_CR[qp_uv-MIN_QP]/6;
  qp_rem_uv = QP_SCALE_CR[qp_uv-MIN_QP]%6;

  // read CBP if not new intra mode
  if (!IS_NEWINTRA (currMB))
  {
    if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB )   currSE.type = SE_CBP_INTRA;
    else                        currSE.type = SE_CBP_INTER;

    dP = &(currSlice->partArr[partMap[currSE.type]]);
    
    if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB)  currSE.mapping = linfo_cbp_intra;
    else                       currSE.mapping = linfo_cbp_inter;

    //coded_block_pattern

    dP->readSyntaxElement(&currSE,dP);
    currMB->cbp = cbp = currSE.value1;
    // Delta quant only if nonzero coeffs
    if (cbp !=0)
    {
      if (IS_INTER (currMB))  currSE.type = SE_DELTA_QUANT_INTER;
      else                    currSE.type = SE_DELTA_QUANT_INTRA;

      dP = &(currSlice->partArr[partMap[currSE.type]]);
      
      currSE.mapping = linfo_se;

      //mb_qp_delta

      dP->readSyntaxElement(&currSE,dP);
      currMB->delta_quant = currSE.value1;
      img->qp= (img->qp-MIN_QP+currMB->delta_quant+(MAX_QP-MIN_QP+1))%(MAX_QP-MIN_QP+1)+MIN_QP;
    }
  }
  else
  {
    cbp = currMB->cbp;
  }
  
  memset(img->cof,0,1536); //1536 = 4*6*4*4*sizeof(int)/sizeof(byte)

  if (IS_NEWINTRA (currMB)) // read DC coeffs for new intra modes
  {
    currSE.type = SE_DELTA_QUANT_INTRA;

    dP = &(currSlice->partArr[partMap[currSE.type]]);
    
    currSE.mapping = linfo_se;

    dP->readSyntaxElement(&currSE,dP);
    currMB->delta_quant = currSE.value1;
    img->qp= (img->qp-MIN_QP+currMB->delta_quant+(MAX_QP-MIN_QP+1))%(MAX_QP-MIN_QP+1)+MIN_QP;

    for (i=0;i<4;i++)
      for (j=0;j<4;j++)
        img->ipredmode[img->block_x+i][img->block_y+j]=DC_PRED;


    readCoeff4x4_CAVLC(img, LUMA_INTRA16x16DC, 0, 0,
                        levarr, runarr, &numcoeff);
    coef_ctr=-1;
    level = 1;                            // just to get inside the loop
    for(k = 0; k < numcoeff; k++)
    {
      if (levarr[k] != 0)                     // leave if len=1
      {
        coef_ctr=coef_ctr+runarr[k]+1;
        i0=SNGL_SCAN[coef_ctr][0];
        j0=SNGL_SCAN[coef_ctr][1];
        img->cof[i0][j0][0][0]=levarr[k];// add new intra DC coeff
      }
    }

    itrans_2(img);// transform new intra DC
  }

  qp_per    = (img->qp-MIN_QP)/6;
  qp_rem    = (img->qp-MIN_QP)%6;
  qp_uv = img->qp + active_pps->chroma_qp_index_offset;
  qp_uv = Clip3(0, 51, qp_uv);
  qp_per_uv = QP_SCALE_CR[qp_uv-MIN_QP]/6;
  qp_rem_uv = QP_SCALE_CR[qp_uv-MIN_QP]%6;
  currMB->qp = img->qp;

  // luma coefficients
  for (block_y=0; block_y < 4; block_y += 2) /* all modes */
  {
    for (block_x=0; block_x < 4; block_x += 2)
    {

      b8 = 2*(block_y/2) + block_x/2;
      for (j=block_y; j < block_y+2; j++)
      {
        for (i=block_x; i < block_x+2; i++)
        {
          ii = block_x/2; jj = block_y/2;
          b8 = 2*jj+ii;

          if (cbp & (1<<b8))  /* are there any coeff in current block at all */
          {
            if (IS_NEWINTRA(currMB))
            {
              readCoeff4x4_CAVLC(img, LUMA_INTRA16x16AC, i, j,
                                  levarr, runarr, &numcoeff);
              start_scan = 1;
            }
            else
            {
              readCoeff4x4_CAVLC(img, LUMA, i, j,
                                  levarr, runarr, &numcoeff);
              start_scan = 0;
            }

			coef_ctr = start_scan-1;
            for (k = 0; k < numcoeff; k++)
            {
              if (levarr[k] != 0)
              {
                coef_ctr             += runarr[k]+1;
                i0=SNGL_SCAN[coef_ctr][0];
                j0=SNGL_SCAN[coef_ctr][1];
                
                currMB->cbp_blk      |= 1 << ((j<<2) + i) ;
                img->cof[i][j][i0][j0]= levarr[k]*dequant_coef[qp_rem][i0][j0]<<qp_per;
              }
            }
          }
          else
          {
            img->nz_coeff[img->current_mb_nr][i][j] = 0;
          }
        }
      }
    }
  }

  /*img->coflumaĳʼϲ*/

  m2 =img->mb_x*2;
  jg2=img->mb_y*2;

  // chroma 2x2 DC coeff
  if(cbp>15)
  {
    for (ll=0;ll<3;ll+=2)
    {
      memset(img->cofu,0,4*sizeof(int));

      readCoeff4x4_CAVLC(img, CHROMA_DC, 0, 0,
                          levarr, runarr, &numcoeff);
      coef_ctr=-1;
      level=1;
      for(k = 0; k < numcoeff; k++)
      {
        if (levarr[k] != 0)
        {
          currMB->cbp_blk |= 0xf0000 << (ll<<1) ;
          coef_ctr=coef_ctr+runarr[k]+1;
          img->cofu[coef_ctr]=levarr[k];
        }
      }

//      for (i=0;i<4;i++)
	  int dequant_cofu=dequant_coef[qp_rem_uv][0][0]<<qp_per_uv;
      img->cofu[0]*=dequant_cofu;
	  img->cofu[1]*=dequant_cofu;
	  img->cofu[2]*=dequant_cofu;
	  img->cofu[3]*=dequant_cofu;

      img->cof[0+ll][4][0][0]=(img->cofu[0]+img->cofu[1]+img->cofu[2]+img->cofu[3])>>1;
      img->cof[1+ll][4][0][0]=(img->cofu[0]-img->cofu[1]+img->cofu[2]-img->cofu[3])>>1;
      img->cof[0+ll][5][0][0]=(img->cofu[0]+img->cofu[1]-img->cofu[2]-img->cofu[3])>>1;
      img->cof[1+ll][5][0][0]=(img->cofu[0]-img->cofu[1]-img->cofu[2]+img->cofu[3])>>1;
    }
  }

  // chroma AC coeff, all zero fram start_scan
  if (cbp<=31)
    for (i=0; i < 4; i++)
      for (j=4; j < 6; j++)
        img->nz_coeff [img->current_mb_nr ][i][j]=0;


  // chroma AC coeff, all zero fram start_scan
  uv=-1;
  if (cbp>31)
  {
    block_y=4;
    for (block_x=0; block_x < 4; block_x += 2)
    {
      for (j=block_y; j < block_y+2; j++)
      {
        jj=j>>1;
        j1=j-4;
        for (i=block_x; i < block_x+2; i++)
        {

          ii=i>>1;
          i1=i&0x0001;

		  readCoeff4x4_CAVLC(img, CHROMA_AC, i, j,
                              levarr, runarr, &numcoeff);
          coef_ctr=0;
          level=1;
          uv++;
          for(k = 0; k < numcoeff;k++)
          {
            if (levarr[k] != 0)
            {
              currMB->cbp_blk |= 1 << (16 + (j1<<1) + i1 + (block_x<<1) ) ;
              coef_ctr=coef_ctr+runarr[k]+1;
              i0=SNGL_SCAN[coef_ctr][0];
              j0=SNGL_SCAN[coef_ctr][1];
              
              img->cof[i][j][i0][j0]=levarr[k]*dequant_coef[qp_rem_uv][i0][j0]<<qp_per_uv;
            }
          }
        }
      }
    }
  }
}


/*!
 ************************************************************************
 * \brief
 *    Copy IPCM coefficients to decoded picture buffer and set parameters for this MB
 *    (for IPCM CABAC and IPCM CAVLC  28/11/2003)
 *
 * \author
 *    Dong Wang <Dong.Wang@bristol.ac.uk>
 ************************************************************************
 */

void decode_ipcm_mb(struct img_par *img)
{
  int i,j;

  Macroblock *currMb = &img->mb_data[img->mb_y*img->PicWidthInMbs + img->mb_x] ;    

  int dy;


  //Copy coefficents to decoded picture buffer
  //IPCM coefficents are stored in img->cof which is set in function readIPCMcoeffsFromNAL()
  dy=1;

  for(i=0;i<16;i++)
    for(j=0;j<16;j++)
      dec_picture->imgY[img->pix_y+i*dy][img->pix_x+j]=img->cof[i/4][j/4][i&0x0003][j&0x0003];

  for(i=0;i<8;i++)
    for(j=0;j<8;j++)
      dec_picture->imgUV[0][img->pix_c_y+i*dy][img->pix_c_x+j]=img->cof[i/4][j/4+4][i&0x0003][j&0x0003];

  for(i=0;i<8;i++)
    for(j=0;j<8;j++)
      dec_picture->imgUV[1][img->pix_c_y+i*dy][img->pix_c_x+j]=img->cof[i/4+2][j/4+4][i&0x0003][j&0x0003];



  //For Deblocking Filter  16/08/2003
  if (currMb->mb_type==IPCM)
    currMb->qp=0;

  //For CAVLC
  //Set the nz_coeff to 16. 
  //These parameters are to be used in CAVLC decoding of neighbour blocks
  for(i=0;i<4;i++)
    for (j=0;j<6;j++)
      img->nz_coeff[img->current_mb_nr][i][j]=16;
}

/*!
 ************************************************************************
 * \brief
 *    decode one macroblock
 ************************************************************************
 */
int decode_one_macroblock(struct img_par *img)
{
  const byte decode_block_scan[16] = {0,1,4,5,2,3,6,7,8,9,12,13,10,11,14,15};
  Macroblock *currMB   = &img->mb_data[img->current_mb_nr];

  
  int i=0,j=0,k,ii=0,jj,j4=0,i4=0;
  int block8x8;   // needed for ABT
  int uv;
  int ioff,joff;

  if(currMB->mb_type==IPCM)
  {
    //copy readed data into imgY and set parameters
    decode_ipcm_mb(img);
    return 0;
  }

  //I16
  if (IS_NEWINTRA (currMB))
  {
    intrapred_luma_16x16(img, currMB->i16mode);
	for (block8x8=0; block8x8<4; block8x8++)
	{
		for (k = block8x8*4; k < block8x8*4+4; k ++)
		{
			i = (decode_block_scan[k] & 3);
			j = ((decode_block_scan[k] >> 2) & 3);
			
			ioff=i<<2;
			i4=(img->block_x+i)<<2;
			
			joff=j<<2;
			j4=(img->block_y+j)<<2;

			itrans   (img,ioff,joff,i,j);      // use DCT transform and make 4x4 block m7 from prediction block mpr
			
			for(ii=0;ii<4;ii++)
				for(jj=0;jj<4;jj++)
					dec_picture->imgY[j4+jj][i4+ii]=img->m7[ii][jj];
		}
	}
	//chroma
	for(uv=0;uv<2;uv++)
	{
		intrapred_chroma(img, uv);
		
		for (j=4;j<6;j++)
		{
			joff=(j-4)<<2;
			j4=img->pix_c_y+joff;

			for(i=0;i<2;i++)
			{
				ioff=i<<2;
				i4=img->pix_c_x+ioff;

				itrans(img,ioff,joff,2*uv+i,j);
				for(ii=0;ii<4;ii++)
					for(jj=0;jj<4;jj++)
						dec_picture->imgUV[uv][j4+jj][i4+ii]=img->m7[ii][jj];
			}
		}
	}
	return 0;
  }

  if(IS_OLDINTRA(currMB))
  {
	for (block8x8=0; block8x8<4; block8x8++)
	{
		for (k = block8x8*4; k < block8x8*4+4; k ++)
		{
			i = (decode_block_scan[k] & 3);
			j = ((decode_block_scan[k] >> 2) & 3);
			
			ioff=i<<2;
			i4=img->block_x+i;
			
			joff=j<<2;
			j4=img->block_y+j;

			if (intrapred(img,ioff,joff,i4,j4)==SEARCH_SYNC)  /* make 4x4 prediction block mpr from given prediction img->mb_mode */
				return SEARCH_SYNC;

			itrans   (img,ioff,joff,i,j);      // use DCT transform and make 4x4 block m7 from prediction block mpr
			i4=i4<<2;
			j4=j4<<2;
			for(ii=0;ii<4;ii++)
				for(jj=0;jj<4;jj++)
					dec_picture->imgY[j4+jj][i4+ii]=img->m7[ii][jj];
		}
	}
	//chroma
	for(uv=0;uv<2;uv++)
	{
		intrapred_chroma(img, uv);
		
		for (j=4;j<6;j++)
		{
			joff=(j-4)<<2;
			j4=img->pix_c_y+joff;
			for(i=0;i<2;i++)
			{
				ioff=i<<2;
				i4=img->pix_c_x+ioff;

				itrans(img,ioff,joff,2*uv+i,j);
				for(ii=0;ii<4;ii++)
					for(jj=0;jj<4;jj++)					
						dec_picture->imgUV[uv][j4+jj][i4+ii]=img->m7[ii][jj];
			}
		}
	}
	return 0;
  }

  int tmp_block[4][4];
  int i1=0,j1=0;
  int vec1_x=0,vec1_y=0;

  int pred;
  int ii0,jj0,ii1,jj1,if1,jf1,if0,jf0;
  int /*mv_mul,*/f1,f2,f3,f4;

  int ref_idx, pred_dir; 

  int  *** mv_array;
  int max_y_cr;

  StorablePicture **list;

  int jf;

  // find out the correct list offsets
  max_y_cr = img->height_cr-1;
  
//  mv_mul=4;
  f1=8;	  f2=7;  f3=64/*f1*f1*/;  f4=f3/2;

  // luma decoding **************************************************

  // get prediction for INTRA_MB_16x16
  for (block8x8=0; block8x8<4; block8x8++)
  {
    for (k = block8x8*4; k < block8x8*4+4; k ++)
    {
      i = (decode_block_scan[k] & 3);
      j = ((decode_block_scan[k] >> 2) & 3);
      
      ioff=i<<2;
      i4=img->block_x+i;
      joff=j<<2;
      j4=img->block_y+j;

      pred_dir = currMB->b8pdir[2*(j>>1)+(i>>1)];
      assert (pred_dir<=2);

      //===== FORWARD/BACKWARD PREDICTION =====
          ref_idx  = dec_picture->ref_idx[LIST_0 + pred_dir][i4][j4];
          mv_array = dec_picture->mv[LIST_0 + pred_dir];
          list     = listX[0+ pred_dir];
          vec1_x = (i4<<4) + mv_array[i4][j4][0];
          vec1_y = (j4<<4) + mv_array[i4][j4][1];

          get_block (ref_idx, list, vec1_x, vec1_y, img, tmp_block);

          if (img->apply_weights)
          {
            for(ii=0;ii<4;ii++)
              for(jj=0;jj<4;jj++)  
                img->mpr[ii+ioff][jj+joff] = Clip1(((img->wp_weight[pred_dir][ref_idx][0] *  tmp_block[ii][jj]+ img->wp_round_luma) >>img->luma_log2_weight_denom)  + img->wp_offset[pred_dir][ref_idx][0] );
          }
          else
          {
			  for(ii=0;ii<4;ii++)
				memcpy(img->mpr[ii+ioff]+joff,tmp_block[ii],4*sizeof(int));
          }

      itrans   (img,ioff,joff,i,j);      // use DCT transform and make 4x4 block m7 from prediction block mpr
       j4<<=2;
	   i4<<=2;
      for(ii=0;ii<4;ii++)
		  for(jj=0;jj<4;jj++)
			  dec_picture->imgY[j4+jj][i4+ii]=img->m7[ii][jj];
    }
  }
  
  // chroma decoding *******************************************************
  for(uv=0;uv<2;uv++)
  {
    for (j=4;j<6;j++)
    {
      joff=(j-4)<<2;
      j4=img->pix_c_y+joff;
      for(i=0;i<2;i++)
      {
        ioff=i*4;
        i4=img->pix_c_x+ioff;
        
        pred_dir = currMB->b8pdir[2*(j-4)+i];
        assert (pred_dir<=2);

        mv_array = dec_picture->mv[LIST_0 + pred_dir];
		list = listX[0+pred_dir];
		for(jj=0;jj<4;jj++)
		{
			jf=(j4+jj)>>1;
			for(ii=0;ii<4;ii++)
			{
                if1=(i4+ii)>>1;
                
                ref_idx   = dec_picture->ref_idx[LIST_0+pred_dir][if1][jf];
                
                i1=(img->pix_c_x+ii+ioff)*f1+mv_array[if1][jf][0];
                j1=(img->pix_c_y+jj+joff)*f1+mv_array[if1][jf][1];
                
                ii0=max (0, min (i1/f1, img->width_cr-1));
                jj0=max (0, min (j1/f1, max_y_cr));
                ii1=max (0, min ((i1+f2)/f1, img->width_cr-1));
                jj1=max (0, min ((j1+f2)/f1, max_y_cr));
                
                if1=(i1 & f2);
                jf1=(j1 & f2);
                if0=f1-if1;
                jf0=f1-jf1;
                
                if (img->apply_weights)
                {
                  pred = (if0*jf0*list[ref_idx]->imgUV[uv][jj0][ii0]+
                          if1*jf0*list[ref_idx]->imgUV[uv][jj0][ii1]+
                          if0*jf1*list[ref_idx]->imgUV[uv][jj1][ii0]+
                          if1*jf1*list[ref_idx]->imgUV[uv][jj1][ii1]+f4)>>6;///f3;
                  img->mpr[ii+ioff][jj+joff] = Clip1(((img->wp_weight[pred_dir][ref_idx][uv+1] * pred  + img->wp_round_chroma)>>img->chroma_log2_weight_denom) + img->wp_offset[pred_dir][ref_idx][uv+1]);
                }
                else
                {
                  img->mpr[ii+ioff][jj+joff]=(if0*jf0*list[ref_idx]->imgUV[uv][jj0][ii0]+
                                              if1*jf0*list[ref_idx]->imgUV[uv][jj0][ii1]+
                                              if0*jf1*list[ref_idx]->imgUV[uv][jj1][ii0]+
                                              if1*jf1*list[ref_idx]->imgUV[uv][jj1][ii1]+f4)>>6;///f3;
                }
			}
		}
        
        itrans(img,ioff,joff,2*uv+i,j);
        for(ii=0;ii<4;ii++)
			for(jj=0;jj<4;jj++)
				dec_picture->imgUV[uv][j4+jj][i4+ii]=img->m7[ii][jj];
      }
    }
  }
  return 0;
}
